/* subj.c:
 *
 * replacement program for the system "from" command.  displays the sender's
 * name as well as the subject of the message (if possible).
 *
 * program written by jim frost.
 *
 * History:
 * 10.19.90  extended the size of lines to BUFSIZ from 81 to fix a problem
 *           where header lines that were exactly 80 characters in length
 *           would cause a premature end-of-header break.
 * 8.21.89   fixed bug where when localhost == domain suffix we could get a
 *           null hostname with preceeding '!' or trailing '@'
 * 8.21.89   added code to edituser() to convert uucp addresses to domain
 *           style addresses if it seems like the right thing to do -- spike
 * 4.11.89   add check to see if TERM environment variable is set.  some
 *           termcaps coredump if it's not.  also fixed bug where tabs
 *           following header labels would cause the label to be ignored.
 * 2.19.89   fix to check for unread mail in BABYL format files -- composer
 * 2.18.89   add ability to check only for unread mail -n flag -- composer
 * 5.4.88    fix bug in header end check for mail type files
 * 4.5.88    add ability to ask termcap for number of columns.
 * 4.4.88    add -d and -w flags.
 * 3.30.88   add -c, -C, -F, -s, -t, and -T options.  add ability to truncate
 *           uppercase hostnames and suffixes.
 * 3.28.88   fix RMAIL mode bug.  if no subject, it would use the subject
 *           from the following message.  also add -v flag so we can
 *           figure out which version we have.
 * 3.25.88   rewrite to clean things up and fix an RMAIL file bug that
 *           caused duplicate subjects to appear.  add ability to delete
 *           the local machine name from the "from" field.  added ability
 *           to recognize UUCP addresses and strip all but original host
 *           and user.
 * 3.24.88   add ability to remove extraneous local suffixes (eg .bu.edu)
 * 3.2.88    added the ability to deal with RMAIL files automatically.
 * 2.23.88   added the ability to read a file (-f flag) and to use other
 *           users' mailfiles.
 * 9.10.87   bug that could cause incorrect subject fixed.  bug that could
 *           cause last message to print twice (if it's incoming) fixed.
 *           string length bug fixed to correctly handle the null at the
 *           end of the string if it was too long.
 * 6.5.87    initial release
*/

#include <stdio.h>
#include <ctype.h>

#define VERSION     "8.21.89"          /* version number */
#define SCRNCOLS    80                 /* columns on screen */
#define TOLEN       15                 /* length of displayed "to" field */
#define FROMLEN     15                 /* length of displayed "from" field */
#define DATELEN     28                 /* length of displayed "date" field */
#define WIDTH       80                 /* default screen width */
#define MAXHOSTNAME 256

#ifdef LOCSUF
char *locsuf= LOCSUF;
#endif

int instr();
void lowstr();
void edituser();
void printfrom();

/* global definitions used for various print modes
 */

int width,            /* window width for line truncating */
    showto,           /* show "To:" field */
    showsub,          /* show "Subject:" field */
    showdate,         /* show "Date:" field */
    fuser,            /* flags printing only when from a particular user */
    tuser,            /* flags printing only when to a particular user */
    read_yet,         /* msg read or not -- 2.18.89 -- composer */
    shownewonly,      /* only display new msgs (in this case, not read) */
                      /*     -- 2.18.89 -- composer */
    newmail;          /* 0, if no new mail, >0 = # of new msgs -- composer */
char fpattern[BUFSIZ],/* pattern that we are looking for messages from */
     tpattern[BUFSIZ],/* pattern that we are looking for messages to */
     cc[BUFSIZ],      /* cc user names used to find -T matches */
     from[BUFSIZ],    /* from user name */
     fto[BUFSIZ],     /* full "to" line used to find -T matches */
     to[BUFSIZ],      /* to user name */
     subject[BUFSIZ], /* subject of message */
     date[BUFSIZ];    /* date the message was sent */

main(argc,argv)
     int argc;
     char *argv[];
{ char mailfile[BUFSIZ], /* mail file name */
       line[BUFSIZ],     /* input buffer */
       tcap[BUFSIZ];     /* termcap entry buffer */
  int  a,
       isfile,  /* we are dealing with a file */
       isuser,  /* we are dealing with a user */
       mailtype,/* type of mail file */
       inheader;/* we are currently in the RMAIL header */
  FILE *f;

  /* initialize everything
   */

  line[BUFSIZ - 1]='\0';
  isfile= 0;
  isuser= 0;
  mailtype= 0;
  fuser= 0;        /* not looking for mail from particular user */
  tuser= 0;        /* not looking for mail to particular user */
  showto= 0;       /* don't show to lines */
  showsub= 1;      /* do show subject lines */
  showdate= 0;     /* don't show date lines */
  shownewonly = 0; /* normally display all msgs -- 2.18.89 -- composer */
  newmail = 0;     /* no new mail yet -- 2.18.89 -- composer */

  /* ask termcap how wide we are
   */

  width= WIDTH; /* default window width */
  if (getenv("TERM") && (tgetent(tcap,getenv("TERM")) > 0))
    width= tgetnum("co");

  /* examine possible command line cases
   */

  for (a= 1; a < argc; a++) {
    if (*argv[a] == '-') {        /* optional switch */
      switch(*(argv[a]+1)) {
      case 'd' : /* display date field */
	showdate= !showdate;
	break;
      case 'f' :
	if (isuser) {
	  printf("%s: you may not specify both a user and a filename\n",
		 argv[0]);
	  exit(1);
	}
	if (isfile) {
	  printf("%s: you may only specify one filename\n",argv[0]);
	  exit(1);
	}
	if (*(argv[a]+2) == '\0')    /* filename in next arg */
	  if (++a < argc)
	    strcpy(mailfile,argv[a]);
	  else {
	    printf("%s: -f flag requires a filename\n",argv[0]);
	    exit(1);
	  }
	else
	  strcpy(mailfile,argv[a]+2); /* filename in this arg */
	isfile= 1;
	break;
      case 'F' :  /* -F -- show only messages matching a from pattern */
	if (fuser) {
	  printf("%s: you may only specify one -F pattern\n",argv[0]);
	  exit(1);
	}
	if (*(argv[a]+2) == '\0')    /* pattern in next arg */
	  if (++a < argc)
	    strcpy(fpattern,argv[a]);
	  else {
	    printf("%s: -F flag requires a pattern\n",argv[0]);
	    exit(1);
	  }
	else
	  strcpy(fpattern,argv[a]+2); /* pattern in this arg */
	lowstr(fpattern);
	fuser= 1;
	break;
      case 'n' :  /* -n -- show only new msgs - 2.18.89 -- composer */
	shownewonly = 1;  /* not a toggle */
	break;
      case 's' :  /* -s -- toggle subject suppression */
	showsub= !showsub;
	break;
      case 't' :  /* -t -- toggle to line printing */
	showto= !showto;
	break;
      case 'T' :  /* -T -- look for messages matching pattern on to line */
	if (tuser) {
	  printf("%s: you may only specify one -T pattern\n",argv[0]);
	  exit(1);
	}
	if (*(argv[a]+2) == '\0')    /* pattern in next arg */
	  if (++a < argc)
	    strcpy(tpattern,argv[a]);
	  else {
	    printf("%s: -T flag requires a pattern\n",argv[0]);
	    exit(1);
	  }
	else
	  strcpy(tpattern,argv[a]+2); /* pattern in this arg */
	lowstr(tpattern);
	tuser= 1;
	break;
      case 'v' :  /* -v -- show version number */
	printf("subj version %s\n",VERSION);
	exit(0);
      case 'w' :  /* set a particular max line length */
	if (*(argv[a]+2) == '\0')    /* number in next arg */
	  if (++a < argc)
	    width= atoi(argv[a]);
	  else {
	    printf("%s: -w flag requires a numeric argument\n",argv[0]);
	    exit(1);
	  }
	else
	  width= atoi(argv[a]+2); /* number in this arg */
	break;
      default:   /* stupid user, tricks are for kids */
	printf("Usage: %s [user] [-d] [-f filename] [-F pattern] [-n] [-s] [-t] [-T pattern] [-v] [-w width]\n",argv[0]);
        exit(1);
      }
    }
    else {                              /* user name */

      /* if the name appears to be a filename and not a user name, do an
       * implied "-f" flag.
       */

      if (instr("/",argv[a])) {
        if (isfile) {
          printf("%s: you may only specify one filename\n",argv[0]);
          exit(1);
        }
        strcpy(mailfile,argv[a]);
        isfile= 1;
      }
      else {
        if (isfile) {
          printf("%s: you may not specify both a username and a filename\n",
                 argv[0]);
          exit(1);
        }
        if (isuser) {
          printf("%s: you may only specify one username\n",argv[0]);
          exit(1);
        }
        strcpy(mailfile,MAILDIR);
        strcat(mailfile,argv[a]);
        isuser= 1;
      }
    }
  }

  if ((!isfile) && (!isuser)) {      /* use current user's name */
    if (getenv("USER")==NULL) {      /* no USER field in environment */
      printf("%s: I don't know who you are (no USER field in environment)\n",
	     argv[0]);
      exit(1);
    }
    strcpy(mailfile,MAILDIR);
    strcat(mailfile,getenv("USER"));
  }

  if ((f=fopen(mailfile,"r"))==NULL) { /* can't read it so die */
    if (isuser || isfile) {            /* report our lack of access */
      perror(mailfile);
      exit(1);
    }
    else
      exit(0);                         /* fine, just exit */
  }

  cc[0]= from[0]= to[0]= subject[0]= date[0]= '\0';
  while (fgets(line,BUFSIZ-1,f)!=NULL) {
    if (line[strlen(line)-1]=='\n') /* strip extraneous newline */
      line[strlen(line)-1]='\0';

    /* determine what type of mail file this is
     */

    if (!mailtype)
      if (!strncmp(line,"BABYL OPTIONS:",14)) { /* RMAIL */
        mailtype= 2;
        inheader= 0;
	read_yet = 1;   /* assume msg has been read if in BABYL format */
                        /*    2.19.89 -- composer */
      }
      else {                                    /* normal */
        mailtype= 1;
	read_yet = 0;   /* assume msg has not been read yet - composer */
      }

    switch (mailtype) {

      /* normal mail mode.  it is assumed that all mail messages
       * start with "From user ...".
       */

      case 1 :
        if ((!strncmp("From ",line,5)) && (strlen(from)==0)) {
          inheader= 1;
          sscanf(&line[5],"%s",from);
          strcpy(subject,"");
        }
        if (inheader) {
          if ((strlen(to) == 0) && (!strncmp("To:",line,3))) {
            strcpy(fto,&line[4]);
            sscanf(&line[4],"%s",to);
          }
          if ((strlen(cc) == 0) && (!strcmp("Cc:",line,3)))
            strcpy(cc,&line[4]);
          if ((strlen(subject)==0) && (!strncmp("Subject:",line,8)))
            strcpy(subject,&line[9]);
          if ((strlen(date) == 0) && (!strncmp("Date:",line,5))) {
            for (a= 6; (line[a] == ' ') || (line[a] == '\t'); a++);
            strcpy(date,&line[a]);
	  }

	  /* added code to check if mail msg has been read or not
	   * 2.18.89 -- composer
	   */ 

          if (!strncmp("Status:",line,7)) {
	    for (a=8; (line[a] != '\0'); a++)
	      if (line[a] == 'R')
		read_yet = 1;
	  }

	  /* when we find the first blank line, it's the end of the header.
	   * print out what we have.  the check for the length of "from"
	   * takes care of multiple printings because from is only defined
	   * at the top of each new message.
	   */

          if (!strcmp("",line)) {
            printfrom();
	    cc[0]= from[0]= to[0]= subject[0]= date[0]= '\0';
            inheader= 0;
	    read_yet = 0;    /* added 2.18.89 -- composer */
          }
        }
        break;

	/* this is for RMAIL mode.  unlike mail, RMAIL preceeds every message
	 * with a header line.  when in RMAIL mode, we look for this header to
	 * indicate the beginning of the header.
	 */

      case 2 : {
	if (!strncmp(line,"\037\014",2)) /* beginning of BABYL header, ^_^L */
	  inheader = 1;
        if (inheader) {
	  if (instr("unseen",line)) /* most likely 'unseen,' is not */
	    read_yet = 0;           /* anywhere else in header */
	                            /* - 2.19.89 - composer */
          if ((strlen(from) == 0) && (!strncmp(line,"From:",5)))
            sscanf(&line[6],"%s",from);
          if ((strlen(to) == 0) && (!strncmp(line,"To:",3))) {
            strcpy(fto,&line[4]);
            sscanf(&line[4],"%s",to);
          }
          if ((strlen(cc) == 0) && (!strcmp("Cc:",line,3)))
            strcpy(cc,&line[4]);
          if ((strlen(subject)==0) && (!strncmp("Subject:",line,8)))
            strcpy(subject,&line[9]);
          if ((strlen(date) == 0) && (!strncmp("Date: ",line,6))) {
            for (a= 6; (line[a] == ' ') || (line[a] == '\t'); a++);
            strcpy(date,&line[a]);
          }
          if (!strcmp(line,"")) {
            printfrom();
	    cc[0]= from[0]= to[0]= subject[0]= date[0]= '\0';
            inheader= 0;
	    read_yet = 1;    /* 2.19.89 -- composer */
          }
        }
        break;
      }

      /* this should never happen but why not trap it anyway
       */

      default :
        printf("%s: This can never happen!\n", argv[0]);
        exit(1);
    }
  }

  /* print out last entry if necessary
   */

  printfrom();
  fclose(f);

  if (shownewonly && !newmail)   /* no new mail  :-(  -- 2.18.89 -- composer */
    printf("No unread mail.\n"); /* used to be "no new mail" */
}

/* look for a substring within a string.  this is used for matching
 * patterns in "from" and "to" lines.
 */

int instr(s,t)
     char *s; /* string to look for */
     char *t; /* string to look in */
{ int a,l;

  l= strlen(s);
  for (a= 0; a <= (strlen(t) - l); a++)
    if (!strncmp(s,t+a,l))
      return(1);
  return(0);
}

/* make a string all lower case
 */

void lowstr(s)
     char *s;
{
  for (; *s; s++)
    if (isupper(*s))
      *s= tolower(*s);
}

/* perform various operations on a user's name to make it prettier
 */

void edituser(u)
     char u[];
{ char locmach[MAXHOSTNAME];
  int a,b,c;
  short domain = 0;

   lowstr(u); /* uncap everything for case independency */

  /* look to see if this is a UUCP address.  if so, remove everything
   * but the last machine name in the path and the user name.
   * 
   * if given hostname!user, see if hostname is domain based (ie it contains
   * at least one '.').  if so, the case, we convert it to user@hostname.
   * BTW, this will change 'hostname!user%host' to 'user%host@hostname',
   * which I think is the rigth way to handle this. -- spike
   *
   * in several places we manually copy strings because strcpy may or may
   * not handle overlapping copies correctly.
   */

    for (a= strlen(u); a > 0; a--)
      if (u[a] == '!') {
        for (a--,c=a; (a > 0) && (u[a] != '!'); a--);

	for (++a,b=a; b <= c;b++)
	  if (u[b] == '.') {
	    domain++;
	    break;
	  }
	if(domain) {
	  u[++c] = '\0'; /* change '!' to NULL to make strcpy happy */
	  strcpy(locmach,&u[a]);        /* extract hostname */
	  for (c++,b= 0; (u[b]= u[c]) != '\0'; c++,b++); /* and put */
	  strcat(u,"@");                /* it back */
	  strcat(u,locmach);            /* together*/

	} else if (a > 0)
	  for (b= 0; (u[b]= u[a]) != '\0'; a++,b++);
        a= 0;
      }

#ifdef LOCSUF
  /* this looks to see if we match a local machine suffix string (such as
   * machine name or domain name) and deletes it if found.
   */

    if ((strlen(u) > strlen(locsuf)) &&
	!strcmp(locsuf,u+(strlen(u)-strlen(locsuf))))
      u[strlen(u)-strlen(locsuf)]= '\0';
#endif

  /* this looks to see if we match a local machine name string and deletes
   * the machine name if found.  note that if gethostname returns the full
   * host name (including suffix) then this won't catch it; if this is the
   * case, this routine should be moved to before the local suffix deletion.
   */

  gethostname(locmach+1,MAXHOSTNAME);
  locmach[0]= '@';
  if ((strlen(u) > strlen(locmach)) &&
       !strcmp(locmach,u+(strlen(u)-strlen(locmach))))
      u[strlen(u)-strlen(locmach)]= '\0';

  /* if local host name is the same as the domain suffix, we can get
   * names like 'name@' instead of just 'name'.  this cleans up.
   */

  if (u[strlen(u) - 1] == '@')
    u[strlen(u) - 1]= '\0';
}

/* this prints out everything in a nice format
 */

void printfrom()
{ int sw; /* max subject length to show */

  if (strlen(from)>0) { /* only print if we actually have a message */

    if (shownewonly && read_yet) /* show new msgs only */
      return; /* not a new msg, has been read - 2.18.89 - composer */
    if (shownewonly)  /* this is for seeing if we actually find any new mail */
      newmail++;      /*    -- 2.18.89 -- composer */

    /* this limits output line size to fit on one line
     */

    sw= width-10-FROMLEN-((showsub && showto)*(TOLEN+1))
                        -((showsub && showdate)*(DATELEN+1));
    if (sw < 0)         /* can't have a negative field width, so this */
      sw= 0;            /* fixes that possibility */

    lowstr(cc);         /* all lower case for case independency */
    lowstr(fto);

    /* here we check to see if the "To:" line had multiple users on it, or if
     * there was a "Cc:" line, to specify that the message was sent to multiple
     * users.
     */

    if ((to[strlen(to)-1] == ',') || (strlen(cc) > 0))
      strcpy(to,"[various]");
    else
      edituser(to);       /* prettify username if not various */
    edituser(from);       /* prettify username */

    if ((tuser && (instr(tpattern,fto) ||     /* look for -T matches in to */
                   instr(tpattern,cc))) ||    /* and cc lines */
        (fuser && instr(fpattern,from)) ||
        ((!tuser) && (!fuser))) {             /* normal mode */
      printf("from: %*.*s",-FROMLEN,FROMLEN,from); /* from whom */
      if (showto)                             /* to whom */
        printf(" %*.*s",-TOLEN,TOLEN,to);
      if (showdate)                           /* when */
        printf(" %*.*s",-DATELEN,DATELEN,date);
      if (showsub)                            /* what it's about */
        if (strlen(subject) > 0) {
          if (strlen(subject) <= sw)
            printf(" \"%s\"",subject);
          else
            printf(" \"%*.*s\"",-sw,sw,subject);
        }
        else
          printf(" [no subject]");
      printf("\n");
    }
  }
}
